#!/usr/bin/env python3

# Copyright (C) 2012-2013  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import errno
import fnmatch
import hashlib
import io
import logging
import os
import shutil
import subprocess
import sys
import time

import xml.sax.saxutils

def make_dirs(path, mode=0o777, only_parents=True, exist_ok=True):
  """Recursively create all parent directories of the given path."""
  if only_parents:
    path = os.path.dirname(path)
  try:
    os.makedirs(path, mode=mode, exist_ok=exist_ok)
  except FileExistsError:
    pass


def touch(path, ref):
  """Emulate the touch command with a reference file."""
  mtime = os.path.getmtime(ref)
  os.utime(path, (mtime,mtime))


def append_missing_slash(path, end=True, start=False):
  """Append final slash to path if missing."""
  if end and not path[-1] == '/':
    path += '/'
  if start and not path[0] == '/':
    path = '/' + path
  return path



MTIME_DEFAULT = 0
MTIME_PRESERVE = -1
MTIME_UPDATE = 1

def save_file(path, content, mtime_mode=MTIME_DEFAULT, mtime=None):
  """
  Write content to a file.

  Parent directories are created as necessary. If the file already exists, it
  will be compared to the content in chunks and only updated from the point
  where it differs, to avoid most unnecessary disk IO.

  The mtime_mode enables some control over mtimes:

  MTIME_DEFAULT
    The mtime is updated if the file is modified.

  MTIME_PRESERVE
    If the file already exists, its mtime ispreserved even if the file is
    modified.

  MTIME_UPDATE
    The mtime is updated even if the file is not modified.

  Returns True if the file was modified, False otherwise.
  """
  if not isinstance(content, bytes):
    content = content.encode('utf-8')
  try:
    st = os.stat(path)
    try:
      bufsize = st.st_blksize
    except AttributeError:
      bufsize = io.DEFAULT_BUFFER_SIZE
    if mtime is None:
      mtime = st.st_mtime
    else:
      mtime_mode = MTIME_PRESERVE

    if st.st_size != len(content):
      with open(path, 'wb') as f:
        f.write(content)
      unmodified = False
    else:
      unmodified = True
      with open(path, mode='r+b', buffering=bufsize) as f:
        content = io.BytesIO(content)
        while unmodified:
          expected = content.read(bufsize)
          buf = f.read(bufsize)
          if expected:
            if expected != buf:
              f.seek(-len(buf), 1)
              f.truncate()
              f.write(expected)
              while expected:
                expected = content.read(bufsize)
                f.write(expected)
              unmodified = False
          elif buf:
            f.truncate()
            unmodified = False
          else:
            break

    if (unmodified and mtime_mode == MTIME_DEFAULT) \
    or mtime_mode == MTIME_PRESERVE:
      os.utime(path, (mtime,mtime))
    elif mtime_mode == MTIME_UPDATE:
      t = time.time()
      os.utime(path, (t,t))

    if not unmodified:
      logging.info('updated {}'.format(path))
      return True

  except OSError as e:
    if e.errno == errno.ENOENT:
      make_dirs(path)
      with open(path, 'wb') as f:
        f.write(content)
      logging.info('updated {}'.format(path))
      return True
    else:
      raise e

  return False




def hash_string(text, hash_type):
  """Return the hash of a string."""
  h = hashlib.new(hash_type)
  h.update(text.encode())
  return h.hexdigest()



def needs_update(target, mtime):
  """Returns True if the target file does not exist or has an mtime older than
  the reference mtime.
  """
  try:
    if os.path.getmtime(target) <= mtime:
      return True
    else:
      return False
  except OSError as e:
    if e.errno == errno.ENOENT:
      return True
    else:
      raise e





def pipe2(cmd, pin=None):
  """Pipe optional input to a command and return the output."""
  p = subprocess.Popen(
    cmd,
    shell=False,
    stdin=subprocess.PIPE,
    stdout=subprocess.PIPE,
    stderr=subprocess.PIPE,
    close_fds=True,
    )
  if pin:
    output, error = p.communicate(input=pin.encode())
  else:
    output, error = p.communicate()
  e = p.wait()
  if e != 0:
    # TODO
    # Replace with logging.error and raise an exception.
    sys.stderr.write('command exited with %d: %s\n' % (e, str(cmd)))
    if pin:
      sys.stderr.write('stdin:\n' + pin + '\n')
    sys.stderr.write('stderr:\n' + error.decode() + '\n')
    sys.exit(1)
  return output.decode()




def format_size(size):
  """Format bytes for tabular display."""
  if size < 0x400:
    return '{:d} B'.format(size)
  else:
    size = float(size) / 0x400
  for prefix in ('KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB'):
    if size < 0x400:
      return '{:0.02f} {}'.format(size, prefix)
    else:
      size /= 0x400
  return '{:0.02f} YiB'.format(size)



RFC_3339_FMT = '%Y-%m-%dT%H:%M:%SZ'

def format_rfc3339_time(t=None):
  """Convert a UNIX time to RFC 3339 format."""
  return time.strftime(RFC_3339_FMT, time.gmtime(t))

ISO_DISPLAY_TIME_FMT = '%Y-%m-%d %H:%M UTC'

def format_iso_display_time(t=None):
  """Convert a UNIX time to RFC 3339 format."""
  return time.strftime(ISO_DISPLAY_TIME_FMT, time.gmtime(t))




def remove_extraneous(path, expected_paths, ignored=None):
  """Remove unexpected files and empty directories from path."""
  if ignored:
    ignored = [os.path.join(path, x.lstrip('/')) for x in ignored]
  else:
    ignored = []
  # TODO: combine these in a single iteration
  for dirpath, dirnames, filenames in os.walk(path, topdown=False):
    for filename in filenames + [append_missing_slash(x) for x in dirnames]:
      fpath = os.path.join(dirpath, filename)
      for pattern in ignored:
        if fnmatch.fnmatch(fpath, pattern):
          break
      else:
        if fpath not in expected_paths:
          try:
            os.unlink(fpath)
            logging.info('removed {}'.format(fpath))
          except OSError as e:
            if e.errno == errno.EISDIR:
              try:
                os.rmdir(fpath)
                logging.info('removed {}'.format(fpath))
              except OSError as e:
                if e.errno == errno.ENOTEMPTY:
                  pass
                else:
                  raise e
            else:
              raise e
          continue
      # Remove broken symlinks.
      if os.path.lexists(fpath) and not os.path.exists(fpath):
        os.unlink(fpath)
        logging.info('removed broken symlink {}'.format(fpath))

  try:
    os.rmdir(path)
    logging.info('removed {}'.format(p))
  except OSError as e:
    if e.errno == errno.ENOTEMPTY:
      pass
    else:
      raise e


def escape(outtype, text):
  """Escape text characters for the given output. This currently only supports
  (X)HTML and XML, and may be incomplete."""
  outtype = outtype.lower()
  if outtype in ('html', 'xhtml', 'xml'):
    return escape_html(text)
  else:
    return text



def escape_html(text):
  """Excape HTML characters."""
#   for c, esc in (
#     ('&', '&amp;'),
#     ('<', '&lt;'),
#     ('>', '&gt;'),
#     ('"', '&quot;'),
#     ("'", '&apos;'),
#   ):
#     text = text.replace(c, esc)
#   return text
  return xml.sax.saxutils.escape(text)


def force_link(target, path, relative=True, target_is_directory=False):
  """Create a symlink at the given path pointing to target."""
  path = path.rstrip('/')
  if relative:
    target = os.path.relpath(target, os.path.dirname(path))
  else:
    target = os.path.abspath(target)
  if os.path.lexists(path):
    if os.path.islink(path) and os.readlink(path) == target:
      return
  try:
    os.unlink(path)
  except IsADirectoryError:
    shutil.rmtree(path)
  except FileNotFoundError:
    pass
  make_dirs(path, only_parents=True, exist_ok=True)
  os.symlink(target, path, target_is_directory=target_is_directory)
