#!/usr/bin/env python3

# Copyright (C) 2012-2013 Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


"""This file contains the defaults used by xyne.archlinux.ca. It is included
both for my own convenience and to provide an example of more advanced usage of
Xac. The documentation needs to be improved but I hope that it is not too
difficult to decipher as is.

It may help to browse the website while looking at the code to see the actual
output.
"""

import colorsys
import errno
import logging
import os.path
import re
from glob import iglob

from xml.dom.minidom import getDOMImplementation

import Xac
import Xac.xml

def hex_to_rgb(color):
  """Convert hexadecimal color strings (#RRGGBB) to RGB 3-tuples."""
  r = float(int(color[1:3],16))/0xff
  g = float(int(color[3:5],16))/0xff
  b = float(int(color[5:7],16))/0xff
  return r,g,b

def hex_to_hsv(color):
  """Convert hexadecimal color strings (#RRGGBB) to HSV 3-tuples."""
  return colorsys.rgb_to_hsv(*hex_to_rgb(color))

def hsv_to_hex(h,s,v):
  """Convert HSV 3-tuples to hexadecimal color strings (#RRGGBB)."""
  r,g,b = colorsys.hsv_to_rgb(h,s,v)
  r = round(0xff * r)
  g = round(0xff * g)
  b = round(0xff * b)
  return "#%02x%02x%02x" % (r,g,b)



def get_default_config_vars(theme_color):
  """Return interpolation variables for the default CSS files.

  Accepts a single hexadecimal color to set the theme. This makes it possible to
  use different colors for different areas of the site.
  """

  # The following code extracts the hue from the given theme color and then uses
  # set saturations and values to generate a theme based on the given hue.
  h,s,v = hex_to_hsv(theme_color)
  padding = 20
  radius = padding
  return {
    'theme_color'              : theme_color,
    'page_color'               : hsv_to_hex(h, 0.0, 0.975),
    'odd_row_color'            : hsv_to_hex(h, 0.0, 0.95),
    'content_font_color'       : '#000000',
    'content_bg_color'         : '#ffffff',
    'content_anchor_color'     : hsv_to_hex(h, s, v*0.8),
    'non_content_font_color'   : hsv_to_hex(h, 0.0, 0.975),
    'non_content_bg_color'     : '#000000',
    'non_content_anchor_color' : hsv_to_hex(h, s*0.8, v),
    'top_links_color'          : hsv_to_hex(h, 0.0, 0.5),

    'border_radius'         : '%dpx' % radius,
    'smaller_border_radius' : '%dpx' % (radius/2),
    'padding'               : '%dpx' % padding,
    'theme_border_width'    : '4px',
    'font_size'             : 'small',

    'monospace_font_family'   : ','.join(["'" + x + "'" for x in (
      'Bitstream Vera Sans Mono',
      'DejaVu Sans Mono',
      'Lucida Console',
      'Monaco',
      'Courier New',
      'Courier',
      )]) + ',monospace',

    'sans_serif_font_family'   : ','.join(["'" + x + "'" for x in (
      'Bitstream Vera Sans',
      'DejaVu Sans',
      'Lucida Sans Unicode',
      'Lucida Grande',
      'Verdana',
      )]) + ',sans-serif',

    'parent_link_prefix'   : '↻ ',
    'navlink_hover_prefix' : '➤ ',
#     'current_area_left'    : '✘',
#     'current_area_right'   : ' '
  }




def get_default_var_delimiters(path):
  """Get default variable delimiters."""
  # Interpolate CSS files.
  if path.endswith('.css'):
    return ('@@@','@@@')
  else:
    return None






def get_default_post_pandoc_handlers(DYN_REAL_PATH, DYN_SRV_PATH, TMPDIR):
  """This function centralizes the creation of the post-pandoc handlers that
  replace the CDATA-like blocks with custom content.
  """
  post_pandoc_handlers = dict()



  # File insertion.
  try:
    import Xac.plugins.insert_file
  except ModuleNotFoundError as e:
    logging.error('FILE handler error: {}'.format(e))
  else:
    post_pandoc_handlers['FILE'] = (
      Xac.plugins.insert_file.insert_file,
      Xac.plugins.insert_file.default_options
    )


  # Syntax highlighting.
  try:
    import Xac.plugins.source_highlight
  except ModuleNotFoundError as e:
    logging.error('CODE/CODEFILE handler error: {}'.format(e))
  else:
    post_pandoc_handlers['CODE'] = (
      Xac.plugins.source_highlight.source_highlight,
      Xac.plugins.source_highlight.default_options
    )
    post_pandoc_handlers['CODEFILE'] = (
      Xac.plugins.source_highlight.source_highlight_file,
      Xac.plugins.source_highlight.default_options
    )



  # LaTeX images.
  try:
    import Xac.plugins.tex2png
  except ModuleNotFoundError as e:
    logging.error('TEX handler error: {}'.format(e))
  else:
    tex_options = Xac.plugins.tex2png.get_default_options(
        DYN_REAL_PATH + '/img/tex',
        DYN_SRV_PATH + '/img/tex',
        tmpdir = TMPDIR + '/tex2png',
        cls = 'xac_tex_math'
    )
    post_pandoc_handlers['TEX'] = (
      Xac.plugins.tex2png.tex2png,
      tex_options
    )



  # Graphviz.
  try:
    import Xac.plugins.graphviz
  except ModuleNotFoundError as e:
    logging.error('GRAPHVIZ handler error: {}'.format(e))
  else:
    graphviz_options = Xac.plugins.graphviz.get_default_options(
        DYN_REAL_PATH + '/img/graphviz',
        DYN_SRV_PATH + '/img/graphviz',
        cls = 'xac_embedded_svg'
    )
    post_pandoc_handlers['GRAPHVIZ'] = (
      Xac.plugins.graphviz.graphviz_svg,
      graphviz_options
    )



  # Babel
  try:
    import Xac.plugins.babel
  except ModuleNotFoundError as e:
    logging.error('INCHI handler error: {}'.format(e))
  else:
    babel_options = Xac.plugins.babel.get_default_options(
        DYN_REAL_PATH + '/img/babel',
        DYN_SRV_PATH + '/img/babel',
        cls = 'xac_embedded_svg'
    )
    post_pandoc_handlers['INCHI'] = (
      Xac.plugins.babel.inchi_svg,
      babel_options
    )



  # ANSI -> HTML conversion with "aha"
  try:
    import Xac.plugins.aha
  except ModuleNotFoundError as e:
    logging.error('ANSI handler error: {}'.format(e))
  else:
    post_pandoc_handlers['ANSI'] = (
      Xac.plugins.aha.convert_ansi,
      Xac.plugins.aha.default_options
    )


  return post_pandoc_handlers






# TODO: catch non-HTML input by checking "typ"
def default_final_pandoc_handler(xac, path, typ, template, text):
  """Handle Pandoc output before saving it."""
  # Remove table column widths because they are incorrectly calculated from raw
  # text that includes invisible elements.
  r = re.compile(r'(?<=\n)<col width="\d+%" />\s*\n')
  text = r.sub('', text)

#   # Highlight code blocks.
#   # This lazily reuses the plugin, wherefore the intermediate object.
#   r2 = re.compile(r'(<pre class="sourceCode ([^"]*?)"><code[^>]*?>(.*?)</code></pre>)', re.S)
#   data = dict()
#   for tag, lang, code in r2.findall(text):
#     if lang == 'xac_code':
#       continue
#     for char, name in (
#       ('&', '&amp;'),
#       ('<', '&lt;'),
#       ('>', '&gt;'),
#       ('"', '&quot;'),
#       ("'", '&apos;'),
#     ):
#       code = code.replace(name, char)
#     data[tag] = lang.strip() + '\n' + code.strip()
#   if data:
#     f, opts = xac.post_pandoc_handlers['CODE']
#     try:
#       text = f(text, data, 'html', opts)
#     except OSError as e:
#       if e.errno != errno.ENOENT:
#         raise e

  # Insert tex2png images.
  r3 = re.compile(r'(<EQ ENV="math">(.*?)</EQ>)', re.S)
  data = dict()
  for tag, tex in r3.findall(text):
    data[tag] = '$' + tex.strip() + '$'
  if data:
    f, opts = xac.post_pandoc_handlers['TEX']
    try:
      text = f(text, None, data, 'html', opts)
    except OSError as e:
      if e.errno != errno.ENOENT:
        raise e
  return text





def get_default_get_pandoc_template(TEMPLATES_DIR):
  """Create a "get_pandoc_template" function that searches for the default
  templates in the given directory.

  This wrapper function accepts the path to the template directory and returns
  a function that searches that directory.
  """
  def get_pandoc_template(xac, path):
    """Return path to template file for given path."""
    if path:
      return os.path.join(TEMPLATES_DIR, 'main.html')
    else:
      return os.path.join(TEMPLATES_DIR, 'listing.html')
  return get_pandoc_template




def get_default_get_pandoc_args(areas, autogen_dir, template_css_dirs, footer=''):
  """Return the default "get_pandoc_args" function.

  areas:
    A list, tuple or other iterable of dictionary objects for defining areas of
    the site. The idea it to be able to give different parts of the site their
    own look and feel via theme colors, logos, etc. Each dictionary contains the
    following:

    path:
      The server path of the area. These are checked in order and the first
      one that is a prefix of the requested path will be used. For example.
      "/foo/bar/" must precede "/foo/" and "/" in the list.

    name:
      The name of the area.

    color:
      The theme color of the area.

    logo:
      The server path to the logo of the area.

  autogen_dir:
    The directory path in which to store the automatically generated CSS files.

  footer_fields:
    An iterable of tuples of field names and values to display in the footer.
  """

  default_upper_links = tuple([a['path'],a['name'],{'style':'color:%s;' % a['color']}] for a in areas)

  impl = getDOMImplementation()


  def get_pandoc_args(
    xac,
    path,
    child_dirnames,
    child_filenames,
    sibling_dirnames,
    sibling_filenames,
    optargs,
    listing=True
  ):
    """
    Return default Pandoc arguments.

    See the Xac class documentation for details.
    """
    srv_path = path.replace(xac.prepub_dir, '', 1)
    if not srv_path:
      srv_path = '/'
    # For the upper links.
    i = 0
    score = 0
    area = None
    for a in areas:
      if srv_path == a['path'].rstrip('/'):
        area = a
        area_index = i
        break
      elif srv_path.startswith(a['path']):
        l = len(a['path'])
        if l > score:
          area = a
          area_index = i
          score = l
      i += 1
    else:
      if not area:
        raise ValueError('no area found for %s' % srv_path)

    color = area['color']
    banner = area['banner']

    src_path = os.path.join(template_css_dirs[0], 'xac.css')
    dest_path = os.path.join(autogen_dir, template_css_dirs[1], 'xac.' + color[1:] + '.css')
    xac_css_path = dest_path.replace(autogen_dir, '', 1)
    # "Static" variable hack, defined below before returning the function,
    # to prevent unnecessary interpolations.
    if color not in get_pandoc_args.colors:
      config_vars = get_default_config_vars(color)
      xac.interpolate_file(src_path, dest_path, config_vars=config_vars)
      get_pandoc_args.colors.add(color)


    default_upper_links[area_index][2]['class'] = 'xac_current_area'
    upper_links = Xac.xml.get_xac_xhtml_nav_bar(
      default_upper_links,
      {'id':'xac_areas'},
      impl=impl
    )
    del default_upper_links[area_index][2]['class']



    server_path = path.replace(xac.prepub_dir, '', 1)
    if listing:
      nav_child_filenames = list()
    else:
      nav_child_filenames = child_filenames
    cwd_nav_panel, par_nav_panel = Xac.xml.get_xac_xhtml_nav_panels(
      xac,
      server_path,
      child_dirnames,
      nav_child_filenames,
      sibling_dirnames,
      sibling_filenames,
      rootname='~',
      classes=['xac_side_panel', 'xac_nav_panel'],
      impl=impl
    )
    nav_panels = cwd_nav_panel + par_nav_panel

    if os.path.isdir(path):
      atom_links = []
      for atom in iglob(os.path.join(path, '*.atom')):
        href = os.path.basename(atom)
        atom_links.append('<link href="%s" type="application/atom+xml" rel="alternate" />' % href)
      if atom_links:
        optargs.extend(('-V', 'atom_links='+'\n'.join(atom_links)))

    common_pandoc_args = [
      '--toc',
      '--base-header-level', '2',
      '-V', 'xac_css_path=' + xac_css_path,
      '-V', 'banner=' + banner,
      '-V', 'upper_links=' + upper_links,
      '-V', 'nav_panels=' + nav_panels,
      '-V', 'footer=' + footer,
      '--indented-code-classes', 'xac_code',
      '--gladtex',
      '--section-divs',
      '--no-highlight',
    ]
    if listing:
      dpath = path.replace(xac.prepub_dir, '', 1)
      markdown_top = '% {0}\n\n'.format('~' + dpath)
      markdown_bottom = ''
      # Suppress size column when listing binaries in server's bin directory.
      bin_dir = srv_path == '/bin' or srv_path == '/cgi-bin'
      listing = Xac.xml.get_xac_xhtml_dir_table(
        xac,
        path,
        child_dirnames,
        child_filenames,
        impl=impl,
        include_size=(not bin_dir)
      )
      return common_pandoc_args + [
        '-V', 'listing=' + listing,
      ] + optargs, markdown_top, markdown_bottom

    else:
      return common_pandoc_args + optargs, None, None

  get_pandoc_args.colors = set()

  return get_pandoc_args

