#!/usr/bin/env python3

# Copyright (C) 2012  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from calendar import timegm
from glob import iglob
from urllib.parse import quote as uquote
from Xac.common import *
from Xac.xml import AtomFeed, parseString
import errno
import json
import os
import re
import shutil
import time



# TODO
# Update this if Pandoc changes
def get_pandoc_identifier(ident):
  """Return a Pandoc identifier for a section header."""
  punctuation = r'''!?'":;`*\/'''
  ident = ident.strip().strip(punctuation).replace(' ', '-').replace('\n', '-').lower().lstrip('0123456789')
  if ident:
    return ident
  else:
    return 'section'




def set_md_date_to_mtime(path):
  """Insert the file's mtime as an RFC 3339 date in the markdown header field."""
  mtime = os.path.getmtime(path)
  date = format_rfc3339_time(mtime)
  cmd = ['pandoc', '-f', 'markdown', '-t', 'markdown', '-s', '-V', 'date='+date, path]
  md = pipe2(cmd)
  return save_file(path, md, mtime_mode=MTIME_PRESERVE)



def get_time_from_md_date(path, unix=True):
  """Extract the time from the date field in the markdown title section.

  If unix is True, a Unix time is returned, otherwise a struct_time object.
  """
  r = re.compile(r'\d{4}-\d\d-\d\dT\d\d:\d\d:\d\dZ')
  with open(path) as f:
    for line in f.readlines():
      m = r.search(line)
      if m:
        obj = time.strptime(m.group(), RFC_3339_FMT)
        if unix:
          return timegm(obj)
        else:
          return obj


def maybe_prefix_mtime(path):
  '''Prefix the mtime to the filename in the format "%Y.%m.%d."'''
  r = re.compile(r'^\d\d\d\d\.\d\d\.\d\d\.')
  fname = os.path.basename(path)
  if not r.match(fname):
    mtime = os.path.getmtime(path)
    prefix = time.strftime('%Y.%m.%d.', time.gmtime(mtime))
    newpath = os.path.join(os.path.dirname(path), prefix + fname)
    os.rename(path, newpath)
    os.utime(newpath, (mtime,mtime))
    return newpath
  else:
    return path



def create_entries(
  dirpath,
  feed,
  page_url,
  entry_func=None,
  markdown=None,
  md_fmt_time = format_iso_display_time,
  md_ext='.markdown',
  xml_ext='.xml',
  xac=None,
):
  """Create a markdown page and a list of atom entry objects from markdown files
  in a given directory.

  The files should contain a title, a date (in RFC 3339 format), a summary
  section, and a content section."""

  if not markdown:
    markdown = ''

  ext_len = len(md_ext)
  reg = re.compile(r'^%(.*?)\n%(.*?)\n(?:%(.*?)\n)?(.*)$', re.S | re.M)

  idents = set()

  for path in sorted(iglob(os.path.join(dirpath, '*'+md_ext)), reverse=True):
    bname = os.path.basename(path)
    pdate = bname[:10].replace('.', '-')
    mtime = os.path.getmtime(path)
    mdate = format_iso_display_time(mtime)
    common_args = ['-s', '-V', 'date='+mdate]
    if xac:
      e_content = xac.pandoc(path, 'html', common_args)
    else:
      e_content = pipe2([
        'pandoc', '-f', 'markdown', '-t', 'html',
        common_args, path
      ])
#     # This ensures regular formatting for the regex.
#     e_markdown = pipe2(['pandoc', '-f', 'markdown', '-t', 'markdown', '-s',
#                       '--base-header-level', '2', path])

    # Using separate calls to Pandoc creates problems with footnotes because
    # it converts inline footnotes to numbered footnotes, leading to overlaps.
    with open(path) as f:
      e_markdown = f.read()


    m = reg.match(e_markdown)
    title = m.group(1)
    if m.group(2).strip():
      authors = [x.strip() for x in m.group(2).split(';') if x]
    else:
      authors = []
    body = m.group(4).replace('\n#', '\n##')




    # TODO
    # Update this if Pandoc changes
    ident = get_pandoc_identifier(title)
    suffix = ''
    i = 1
    while ident + suffix in idents:
      suffix = '-%d' % i
      i += 1
    ident += suffix
    idents.add(ident)
    entry_url = os.path.join(page_url, '#' + uquote(ident))


    path_year, path_month, path_rest = bname.split('.', 2)
    if page_url.endswith("%s/%s" % (path_year, path_month)):
#       archive_url = os.path.join(page_url, '#' + uquote(ident))
      archive_url = page_url
    else:
#       archive_url = os.path.join(page_url, path_year, path_month, '#' + uquote(ident))
      archive_url = os.path.join(page_url, path_year, path_month)

    markdown += '''# %s
------------- -----
Created       %s
Last Modified %s
Archive Link  <%s>
------------- -----

%s

''' % (title, pdate, mdate, archive_url, body)

    entry = feed.create_entry(os.path.join(feed.id, '#' + uquote(ident)), title, mtime)
    for a in authors:
      entry.add_person('author', a)
    i = e_content.index('<html')
    html = e_content[i:]
    body_re = re.compile(r'<html[^>]+xmlns="(.+?)".*(<body[^>]*>.*?</body>)', re.S)
    m = body_re.search(e_content)
    if m:
      xmlns = m.group(1)
      body = m.group(2)
      body = body.replace('<body', '<div xmlns="%s"' % xmlns, 1)
      body = body.replace('</body>', '</div>', 1)
      entry.add_element('content', attr={'type':'xhtml'}, pre=body)
#     html = parseString(e_content)
#     xmlns = html.documentElement.getAttribute('xmlns')
#     body = html.getElementsByTagName('body')[0]
#     body.tagName = 'div'
#     body.setAttribute('xmlns', xmlns)
#     entry.add_element('content', attr={'type':'xhtml'}, node=body)
    entry.add_element('link', {'href':entry_url})
    if archive_url:
      entry.add_element('link', {'href':archive_url, 'rel':'via'})

    # Entry data can be supplemented with optional XML files, e.g. to save a
    # list of contributors alongside the contents.
    try:
      with open(path[:-ext_len] + xml_ext) as f:
        entry.add_xml(f)
    except IOError as e:
      if e.errno != errno.ENOENT:
        raise e

    if entry_func:
      entry_func(entry)

  return markdown


def create_feed(
  dirpath,
  iD,
  title,
  page_url,
  updated=None,
  entry_func=None,
  md_authors = None,
  md_fmt_time = format_iso_display_time,
  md_ext='.markdown',
  xml_ext='.xml',
  xac=None,
):
  """Create a markdown page and an atom feed from markdown files in a given
  directory."""
  if not updated:
    updated = time.time()

  markdown = '% ' + title + '\n%'
  if md_authors:
    markdown += ' ' + '; '.join(md_authors)
  markdown += '\n% ' + md_fmt_time(updated) + '\n\n'


  feed = AtomFeed(iD, title, updated)
  markdown = create_entries(
    dirpath,
    feed,
    page_url,
    entry_func=entry_func,
    markdown=markdown,
    md_fmt_time=md_fmt_time,
    md_ext=md_ext,
    xml_ext=xml_ext,
    xac=xac,
  )
  return markdown, feed



def get_publish_time(path):
  """Parse the prefixed publishing time from the file path and return a time
  struct."""
  return time.strptime(os.path.basename(path)[:10], '%Y.%m.%d')


# TODO
# Add a strftime parameter for arbitrary subpath creation. Other functions will
# need to be modified as well (time prefix functions, the index page generator)
def archive(dpath, ttl= 31 * 24 * 60 * 60, keep=5):
  """Sort entries in a directory into subdirectories by year and month. This
  makes it possible to provide "permanent" links to entries Timestamps are
  prefixed to the names to make them independent of local modification times.

  A number of recent entries will be preserved as symlinks in the root directory
  depending on time and a fixed number to keep.

  ttl: time to live, i.e. how long to leave them in the main dir

  keep: minimum number of entries to keep in the dir, to prevent inactive dirs
        from being cleared
  """

  symlinks = dict()
  abs_dpath = os.path.abspath(dpath)
  for item in iglob(os.path.join(dpath, '*.markdown')):
    # Ignore symlinks that point to archives in the current dir with the same
    # name.
    try:
      link_target = os.path.abspath(os.path.join(dpath, os.readlink(item)))
      name = os.path.basename(item)
      if link_target.startswith(abs_dpath) \
      and link_target.endswith(name):
        symlinks[name] = link_target
        continue
    except OSError as e:
      if e.errno != errno.EINVAL:
        raise e

    p = maybe_prefix_mtime(item)
    year, month, rest = os.path.basename(p).split('.', 2)
    new_path = os.path.join(
      dpath,
      year,
      month,
      os.path.basename(p)
    )
    make_dirs(new_path)
    shutil.move(p, new_path)

  for target, name in get_index_page_entries(dpath, ttl, keep):
    path = os.path.join(dpath, name)
    try:
      current_target = symlinks[name]
      del symlinks[name]
      if target == current_target:
        continue
      else:
        os.unlink(path)
    except KeyError:
      pass
    os.symlink(os.path.relpath(target, dpath), path)

  for name in symlinks:
    path = os.path.join(dpath, name)
    os.unlink(path)




def get_index_page_entries(dpath, ttl= 31 * 24 * 60 * 60, keep=5):
  """Sort entries in a directory into subdirectories by year and month if older
  than a given time to live.

  ttl: time to live, i.e. how long to leave them in the main dir

  keep: minimum number of entries to keep in the dir, to prevent inactive dirs
        from being cleared
  """
  now = time.time()
  struct_now = time.gmtime(now)
  year = struct_now.tm_year
  month = struct_now.tm_mon
  current_entries = 0
  ttl_timestamp = time.strftime('%Y.%m.%d', time.gmtime(now - ttl))
  dpath = os.path.abspath(dpath)

  y_glob = os.path.join(dpath, '[0-9][0-9][0-9][0-9]')
  for y in sorted(iglob(y_glob), reverse=True):
    m_glob = os.path.join(y, '[0-9][0-9]')
    for m in sorted(iglob(m_glob), reverse=True):
      f_glob = os.path.join(m, '*.markdown')
      for f in sorted(iglob(f_glob), reverse=True):
        entry = os.path.basename(f)
        # A benefit of ISO timestamps: string comparison.
        if entry > ttl_timestamp or keep > current_entries:
          current_entries += 1
          yield f, entry
        else:
          break
      else:
        continue
      break
    else:
      continue
    break


def get_news_dirs(path):
  """Yield main directory and all year/month subdirectories."""
  yield path, None, None
  for y_path in sorted(iglob(os.path.join(path, '[0-9][0-9][0-9][0-9]')), reverse=True):
    year = os.path.basename(y_path)
    for m_path in sorted(iglob(os.path.join(y_path, '[0-9][0-9]')), reverse=True):
      month = os.path.basename(m_path)
      yield m_path, year, month



def prev_cur_next(foo):
  """Yield previous, current and next item in an iterable."""
  a = None
  b = next(foo)
  try:
    c = next(foo)
    yield a, b, c
  except StopIteration:
    yield a, b, None
    raise StopIteration
  for n in foo:
    a = b
    b = c
    c = n
    yield a, b, c
  yield b, c, None




def gen_news_feed(
  src_path, dest_path, srv_path, srv_url='/',
  author=None, entry_func=None, title='News', last_update=0,
  ttl=31 * 24 * 60 * 60, keep=5, xac=None):
  """Generate a news feed and index file from markdown files in a directory.

  src_path:
    The directory containing the files.

  dest_path:
    The root directory in which to generate the index pages and atom feeds.

  srv_path:
    The relative path of the directory on the server.

  srv_url:
    The URL to use for links in the feed.

  author:
    The author of the feed.

  entry_func:
    A function that takes an entry object and modifies it, e.g. by adding custom
    links or contributors).

  title:
    The title of the feed.

  last_update:
    The time of the last update, used for determining if an update is necessary.

  ttl:
    The time to wait before moving entries into subdirectories.

  keep:
    The minimum number of entries to keep in the directory, regardless of time.

  xac:
    A Xac object.
  """

  archive(src_path, ttl, keep)
  atom_name = title.lower() + '.atom'
  for prev, cur, next in prev_cur_next(get_news_dirs(src_path)):
    dir_path, year, month = cur
    news_dir = dir_path.replace(src_path, dest_path, 1)
    markdown_path = os.path.join(news_dir, 'index.markdown')
    atom_path = os.path.join(news_dir, atom_name)

    srv_subpath = os.path.normpath(os.path.join(
      srv_path,
      os.path.relpath(dir_path, src_path)
    ))
    srv_suburl = os.path.join(
      srv_url,
      srv_subpath
    )


    updated = 0
    for p in iglob(os.path.join(dir_path, '*.markdown')):
      mtime = os.path.getmtime(p)
      updated = max(updated, mtime)

    updated = max(updated, os.path.getmtime(dir_path))
    latest_news = max(updated, last_update)

    if needs_update(markdown_path, latest_news) \
    or needs_update(atom_path, latest_news):

      page_url = srv_suburl + '/'
      atom_url = atom_path.replace(dest_path, srv_suburl, 1)

      if not entry_func:
        def entry_func(entry):
          if author:
            entry.add_person('author', author)
#           entry.add_element('link', {'href':page_url})

      if year:
        page_title = "%s/%s/%s" % (title, year, month)
      else:
        page_title = title

      markdown, atom = create_feed(
        dir_path,
        srv_suburl + '/',
        page_title,
        srv_suburl,
        updated=updated,
        entry_func=entry_func,
        xac=xac,
      )
      if author:
        atom.add_person('author', author)
      atom.add_element('link', {'href':page_url})
      atom.add_element('link', {'href':atom_url,'rel':'self'})

      if prev:
        prev_atom_url = os.path.join(
          srv_url,
          os.path.normpath(os.path.join(
            srv_subpath,
            os.path.relpath(prev[0], dir_path),
            atom_name
          ))
        )
        atom.add_element('link', {'href':prev_atom_url,'rel':'previous'})

      if next:
        next_atom_url = os.path.join(
          srv_url,
          os.path.normpath(os.path.join(
            srv_subpath,
            os.path.relpath(next[0], dir_path),
            atom_name
          ))
        )
        atom.add_element('link', {'href':next_atom_url,'rel':'next'})

      modified = False
      if save_file(markdown_path, markdown, mtime_mode=MTIME_UPDATE):
        modified = True
      if save_file(atom_path, str(atom), mtime_mode=MTIME_UPDATE):
        modified = True
      if xac and modified:
        xac.modified = True
      return modified
