#!/usr/bin/env python3

# Copyright (C) 2012-2016  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from Xac.common import *
from fnmatch import fnmatch
from glob import iglob

import Xac
import errno
import os
import re
import shutil
import sys
import time
import types

HASH = 'sha1'
MD_EXT = '.markdown'
MD_EXT_LEN = len(MD_EXT)
MDMETA_EXT = '.mdmeta'
MDMETA_EXT_LEN = len(MDMETA_EXT)


################################################################################
# Xac Class
################################################################################

class Xac():
#   PREPUB_DIR = os.getenv('XAC_PREPUB_DIR', default='prepub')
#   PUB_DIR = os.getenv('XAC_PUB_DIR', default='pub')
#   TEMPLATES_DIR = os.getenv('XAC_TEMPLATES_DIR', default="templates")

  def __init__(
    self,
    file_dirs,
    get_pandoc_template,
    get_pandoc_args,
    config_vars={},
    get_var_delimiters=lambda x: None,
    watchlist=[],
    index_file='index',
    post_pandoc_handlers={},
    final_pandoc_handler=lambda v,w,x,y,z: z,
    format_display_date=format_iso_display_time,
    prepub_dir='prepub',
    pub_dir='pub',
    ignored=None,
  ):
    '''Initialize the configuration object.

    get_pandoc_template:
      A method that must accept a relative file path to a markdown file in the
      prepub directory and return a path to a pandoc output template. If the
      argument is None, it should return a template for displaying a directory
      listing.

      For example,

          get_pandoc_template(None) ->
          /path/to/listing.html

    get_pandoc_args:
      A method that must accept

          * path: prepub path of target file
          * child_dirnames: list of child directories
          * child_filenames: list of child files
          * sibling_dirnames: list of sibling directories
          * sibling_filenames: list of sibling files
          * optargs: a list of optional arguments for pandoc that may be
            overridden
          * listing: a boolean indicating if the path is a directory listing

      and return a tuple containing a list of pandoc arguments, optional
      markdown to include at the top of the markdown input, and optional input
      to include at the bottom of the markdown input. The middle will contain
      content read from files. For example,
        (
         ['-V', 'footer=<some footer>'],
         "%/some/path\n\n",
         "#Directory Listing\n\n"
        )
      would create the page title "/some/path" and append a directory listing
      section to the end, which may be followed in the templated by said
      listing.

      The arguments must include any variable-value pairs required for the
      template returned by get_pandoc_template above. The optional pandoc input is
      only used the "listing" argument is true. All converted markdown files
      will acquire the same extension as the template.

    config_vars:
      A dictionary of variable:value pairs for file interpolations.

    get_var_delimiters:
      A function that accepts a file path and either returns a tuple of variable
      delimiters for interpolation or None if no interpolation should occur. For
      example, the function could use fnmatch to recognize css files with
      '*.css" and return ('@@@', '@@@'). All variables in the css files such as
      "@@@foo@@@" would then be replaced with their matching values in the
      config_vars dictionary.

    file_dirs:
      A list of tuples that specify files to publish. Each tuple consists of a
      path to a directory of files and the *relative* path on the server where
      they should be published, e.g. (/path/to/image/files, img).

    watchlist:
      A list of file glob patterns. Files that match the glob patters will have
      their mtimes added to a list that will then be compared against published
      files when decided whether an update is required.

      This list should contain patterns that match all files that affect output.

    index_file:
      The name of the server directory index files, without any extension, e.g.
      "index".

    post_pandoc_handlers:
      A dictionary of tag:handler pairs for processing files after they have
      been passed through pandoc. The handler must accept the current text,
      the source path of the text (i.e. the markdown file path), a dictionary
      of data chunks, the output type, and a dictionary of additional options.
      The plugin code itself should make this clearer.

    final_pandoc_handler:
      A method to handle any final processing before the converted markdown
      file is written to the pub directory. It accepts the Xac object, input
      file path, the input type (e.g. 'html'), the template file path, and the
      output to process. It should return the final output.

    format_display_date:
      A function that accepts a UNIX time and returns a formatted string.

    prepub_dir:
      The path of the prepublication directory in which files will be staged
      before publication.

    pub_dir:
      The publication directory where the server files will be saved.

    ignored:
      Glob patterns of paths to ignore. These can be used to include unmanaged
      content on the site.
    '''

    self.prepub_dir = prepub_dir
    self.pub_dir = pub_dir

    self.get_pandoc_template = types.MethodType(get_pandoc_template, self)
    self.get_pandoc_args = types.MethodType(get_pandoc_args, self)
    self.final_pandoc_handler = types.MethodType(final_pandoc_handler, self)

    self.config_vars = config_vars
    self.get_var_delimiters = get_var_delimiters
    if not file_dirs:
      raise ValueError("no file directories")
    self.file_dirs=file_dirs
    self.post_pandoc_handlers = post_pandoc_handlers

    # TODO
    # Make this an argument. Wtf was I thinking? It was years ago, but still.
    watchlist_files = set([sys.argv[0]])
    for pattern in watchlist:
      for path in set(iglob(pattern)):
        watchlist.add(path)
    for mod in sys.modules.values():
      try:
        watchlist_files.add(mod.__file__)
      except AttributeError:
        pass
    watchlist_files = set((os.path.abspath(x) for x in watchlist_files if x))

    self.watchlist = watchlist_files
    t = 0
    for x in watchlist_files:
      try:
        t = max(t, os.path.getmtime(x))
      except FileNotFoundError:
        pass
    self.watchlist_mtime = t

    self.index_file = index_file
    self.format_display_date = format_display_date

    self.modified = False

    if ignored:
      ignored = set(ignored)
    else:
      ignored = set()
    self.ignored = ignored



  def map_markdown_filenames(self, path, filenames, same_length=False):
    '''Map filenames to the list as they will appear on the server.

    This just hides index pages and changes MD_EXT extensions to the
    matching template extension.

    Path is the parent directory of the files in the prepub dir.
    '''
    for filename in filenames:
      if filename.endswith(MD_EXT):
        if filename[:MD_EXT_LEN] == self.index_file:
          # yield None to preserve length and enable zipping
          if same_length:
            yield None
        else:
          template = self.get_pandoc_template(os.path.join(path, filename))
          root, ext = os.path.splitext(template)
          yield filename[:-9] + ext
      else:
        yield filename



  def pre_pandoc(self, text, inpath, blocks=None, escaped_tags=None):
    '''Preprocess content before passing it to Pandoc.

    This function recognizes CDATA-like blocks (e.g. <![TAG[DATA]TAG]>) and hides
    them from Pandoc using hash tags. The data is then reinserted after Pandoc has
    been run, with optional further processing by post_pandoc and the configurable
    handlers.
    '''

    if blocks == None:
      blocks = {}

    if escaped_tags == None:
      escaped_tags = dict()

    # Trying to capture "!*" breaks the expression. I gave up trying to
    # understand why.
    r = re.compile(r'(<!!*\[([^\]\n]+)\[(.+?)]\2]>)', re.S)
    replaced = text
    for insert, tag, data, in r.findall(text):
      key = hash_string(insert, HASH)
      if insert[2] == '!':
        # More than one exclamation point escapes the block, and one is removed
        # before reinserting the block. To show n exclamation points, n+1 are
        # needed. This can thus handle all possible cases.

        # Here the tags are replaced with the hash to hide this block when
        # replacing the represented block. For example, if
        # <!![FOO[...]FOO]>
        # were immediately replaced with
        # <![FOO[...FOO]>
        # then that block would itself be replaced when intesting the contents
        # of a separate
        # <![FOO[...FOO]>.
        i = 3
        while insert[i] == '!':
          i += 1
        replaced = replaced.replace(
          insert,
          '<{0}[{1}[{2}]{1}]>'.format(insert[2:i], key, data)
        )
        escaped_tags[key] = tag
      else:
        replaced = replaced.replace(insert, key)
        try:
          blocks[tag][key] = data
        except KeyError:
          blocks[tag] = {key:data}
    return replaced, blocks, escaped_tags




  def post_pandoc(self, text, inpath, blocks, escaped_tags, outtype):
    '''Handle post-Pandoc replacements.'''
    for tag, data in blocks.items():
      try:
        f, opts = self.post_pandoc_handlers[tag]
      except KeyError:
        pass
      else:
        # This replaces the previous inserted tags with the plugin output then
        # skips to the next item.
        try:
          text = f(text, inpath, data, outtype, opts)
          continue
        except OSError as e:
          if e.errno != errno.ENOENT:
            raise e
      logging.warning('failed to replace "{}" blocks'.format(tag))
      # This is only invoked if the plugin failed.
      for hashtag, content in data.items():
        block = '<![{0}[{1}]{0}]>'.format(tag, content)
        block = escape(outtype, block)
        if outtype[-4:] == 'html':
          block = r'<pre>{}</pre>'.format(block)
        text = text.replace(hashtag, block)
    for key, tag in escaped_tags.items():
      text = text.replace(key, tag)
    return text





  def pandoc(self, inpath, outtype, pandoc_args, markdown_top=None, markdown_bottom=None):
    index_file = self.index_file + MD_EXT
    if markdown_top:
      markdown = markdown_top
    else:
      markdown = ''
    if os.path.basename(inpath) == index_file:
      try:
        with open(inpath) as fi:
          markdown += fi.read() + '\n\n'
      # Directory listings are shown when no index file is present.
      except FileNotFoundError:
        pass
      dirpath = os.path.dirname(inpath)
      for m in sorted(iglob(os.path.join(dirpath, '*'+MDMETA_EXT))):
        with open(m) as fm:
          markdown += fm.read() + '\n\n'
    else:
      with open(inpath) as fi:
        markdown += fi.read() + '\n\n'
    if markdown_bottom:
      markdown += markdown_bottom + '\n\n'
    text, blocks, escaped_tags = self.pre_pandoc(markdown, inpath)
    pandoc = [
      'pandoc',
      '-f', 'markdown',
      '-t', outtype
      ] + pandoc_args
    text = pipe2(pandoc, text)
    # Plugins usually create files. Assume modification. The plugin will only
    # be re-invoked if the calling markdown is modified again.
#       self.modified = True
    return self.post_pandoc(text, inpath, blocks, escaped_tags, outtype)





  def interpolate_file(self, src_path, dest_path, config_vars=None, delimiters=None):
    '''Interpolate source file and write results to destination file.'''
    if config_vars == None:
      config_vars = self.config_vars

    if not delimiters:
      delimiters = self.get_var_delimiters(src_path)

    if delimiters:
      left = delimiters[0].encode('utf-8')
      right = delimiters[1].encode('utf-8')
      ref_mtime = max(self.watchlist_mtime, os.path.getmtime(src_path))
      if needs_update(dest_path, ref_mtime):
        with open(src_path, 'rb') as f:
          text = f.read()
        original = text
        for var, val in config_vars.items():
          text = text.replace(left + var.encode('utf-8') + right, val.encode('utf-8'))
        if text != original:
          # Remove symlinks before writing to avoid overwriting original
          # file.
          if os.path.islink(dest_path):
            try:
              os.unlink(dest_path)
              with open(dest_path, 'wb') as f:
                f.write(text)
              return
            except OSError as e:
              if e.errno == errno.ENOENT:
                pass
              else:
                raise e
          if save_file(dest_path, text, mtime_mode=MTIME_UPDATE):
            self.modified = True
          return
      else:
        # No update required.
        return
    # Symlink target path if no changes are required.
    target = os.path.relpath(src_path, os.path.dirname(dest_path))
    try:
      if os.readlink(dest_path) == target:
        return
    except OSError as e:
      # Remove regular files.
      if e.errno == errno.EINVAL:
        try:
          os.unlink(dest_path)
        except OSError as e:
          if e.errno == errno.EISDIR:
            shutil.rmtree(dest_path)
          else:
            raise e
      elif e.errno == errno.ENOENT:
        pass
      else:
        raise e
    try:
      os.unlink(dest_path)
    except OSError as e:
      if e.errno != errno.ENOENT:
        raise e
    make_dirs(dest_path)
    os.symlink(target, dest_path)



  def ignore(self, srv_path):
    '''Check if given server path should be ignored.'''
    if self.ignored:
      for pattern in self.ignored:
        if fnmatch(srv_path, pattern):
          return True
    return False



  def prepublish(self):
    '''Prepublish files to the prepub directory in preparation for publishing.'''
    expected_paths = set()

    for real_path, rel_path in self.file_dirs:
      if not os.path.isdir(real_path):
        continue

      real_path = append_missing_slash(real_path)

      if rel_path and rel_path != '/':
        rel_path = append_missing_slash(rel_path, start=True)
        prepub_dir = self.prepub_dir.rstrip('/') +  rel_path
      else:
        rel_path = '/'
        prepub_dir = append_missing_slash(self.prepub_dir)

      if self.ignore(rel_path):
        force_link(real_path, prepub_dir, target_is_directory=True)
        continue

      for dirpath, dirnames, filenames in os.walk(real_path, followlinks=True):
        ds = list(dirnames)
        dirnames.clear()
        for d in ds:
          src_path = os.path.join(dirpath,d)
          srv_path = src_path.replace(real_path, rel_path, 1)
          if self.ignore(srv_path):
            dest_path = src_path.replace(real_path, prepub_dir, 1)
            force_link(src_path, dest_path, target_is_directory=True)
          else:
            dirnames.append(d)
        for filename in filenames:
          src_path = os.path.join(dirpath, filename)
          srv_path = src_path.replace(real_path, rel_path, 1)
          # The destination path within the prepub directory.
          dest_path = src_path.replace(real_path, prepub_dir, 1)
          if self.ignore(srv_path):
            force_link(src_path, dest_path)
          else:
            # Track current paths to recognize old paths when walking the prepub dir.
            expected_paths.add(dest_path)
            self.interpolate_file(src_path, dest_path)

    remove_extraneous(self.prepub_dir, expected_paths, self.ignored)





  def publish_html(self):
    '''Publish the contents of the prepub directory.

    This will convert markdown files to (X)HTML with Pandoc along with any
    specified pre- and post-processing. Non-markdown files will be symlinked
    into the pub directory.
    '''
    self.prepublish()
    pre_watchlist_mtime = self.watchlist_mtime

    # track current paths so we can remove extraneous files from the pub dir later
    expected_paths = set()

    # dictionary of siblings for generating nav tree
    siblings = {'/' : ((),())}
    for dirpath, dirnames, filenames in os.walk(self.prepub_dir):
      watchlist_mtime = max(pre_watchlist_mtime, os.path.getmtime(dirpath))
      if dirpath != self.prepub_dir:
        watchlist_mtime = max(watchlist_mtime, os.path.getmtime(os.path.dirname(dirpath)))
      offset = len(self.prepub_dir)

      # Ignore paths.
      xs = list(dirnames)
      ignored_dirnames = list()
      dirnames.clear()
      for x in xs:
        src_path = os.path.join(dirpath, x)
        srv_path = append_missing_slash(src_path[offset:])
        if self.ignore(srv_path):
          dest_path = src_path.replace(self.prepub_dir, self.pub_dir, 1)
          force_link(src_path, dest_path, target_is_directory=True)
          ignored_dirnames.append(x)
        else:
          dirnames.append(x)

      xs = list(filenames)
      ignored_filenames = list()
      allpaths = [os.path.join(dirpath, x) for x in filenames]
#       mdmetapaths = [p for p in allpaths if p.endswith(MDMETA_EXT)]
      filenames.clear()
      for x in xs:
        if x.endswith(MDMETA_EXT):
          continue
        src_path = os.path.join(dirpath, x)
        srv_path = src_path[offset:]
        if self.ignore(srv_path):
          dest_path = src_path.replace(self.prepub_dir, self.pub_dir, 1)
          force_link(src_path, dest_path, target_is_directory=False)
          ignored_filenames.append(x)
        else:
          filenames.append(x)

      index_file = self.index_file + MD_EXT
      nav_filenames = filenames + ignored_filenames
      nav_dirnames = dirnames + ignored_dirnames

      server_path = dirpath[offset:]
      if not server_path:
        server_path = '/'
      dir_sibling_dirs, dir_sibling_files = siblings[os.path.dirname(server_path)]
      siblings[server_path] = (nav_dirnames, nav_filenames)

      # Insert directory listing when there is no index page.
      if not index_file in filenames and (filenames or dirnames):
        is_listing = True
        filenames.append(index_file)
      else:
        is_listing = False

      for filename in filenames:
        fullname = os.path.join(dirpath, filename)
        pubname = fullname.replace(self.prepub_dir, self.pub_dir,1)
        make_dirs(pubname)

        if filename.endswith(MD_EXT):
          rel_path = fullname.replace(self.prepub_dir + '/', '', 1)
          if filename == index_file and is_listing:
            template = self.get_pandoc_template(None)
          else:
            template = self.get_pandoc_template(rel_path)
          try:
            content_mtime = os.path.getmtime(fullname)
          except FileNotFoundError as e:
            if filename != index_file:
              raise e
            else:
              content_mtime = 0
          ref_mtime = max(watchlist_mtime, os.path.getmtime(template), content_mtime)
          root, ext = os.path.splitext(template)
          pubname = pubname[:-MD_EXT_LEN] + ext
          expected_paths.add(pubname)

          if filename == index_file:
            for fpath in allpaths:
              t = os.path.getmtime(fpath)
              ref_mtime = max(ref_mtime, t)
              content_mtime = max(content_mtime, t)
            optargs = [
              '-V', 'date=' + self.format_display_date(content_mtime),
              '-V', 'date_rfc3339=' + format_rfc3339_time(content_mtime),
            ]
            pandoc_args, markdown_top, markdown_bottom = self.get_pandoc_args(
              dirpath,
              nav_dirnames, nav_filenames,
              dir_sibling_dirs, dir_sibling_files,
              optargs, listing=is_listing
            )

          else:
            optargs = [
              '-V', 'date=' + self.format_display_date(content_mtime),
              '-V', 'date_rfc3339=' + format_rfc3339_time(content_mtime),
            ]
            pandoc_args, markdown_top, markdown_bottom = self.get_pandoc_args(
              os.path.join(dirpath, filename[:-9]+ext),
              [], [],
              dirnames, filenames,
              optargs, listing=False
            )
          # Don't update if newer.
          if not needs_update(pubname, ref_mtime):
            continue
          pandoc_args = ['--template', template] + pandoc_args
          text = self.pandoc(
            fullname, 'html', pandoc_args,
            markdown_top=markdown_top,
            markdown_bottom=markdown_bottom,
          )
          text = self.final_pandoc_handler(fullname, 'html', template, text)
          if save_file(pubname, text, mtime_mode=MTIME_UPDATE):
            self.modified = True


        else:
          expected_paths.add(pubname)
          target = os.path.relpath(fullname, os.path.dirname(pubname))
          if os.path.lexists(pubname):
            if os.path.islink(pubname) and os.readlink(pubname) == target:
              continue
            else:
              try:
                os.unlink(pubname)
              except OSError as e:
                if e.errno == errno.EISDIR:
                  shutil.rmtree(pubname)
                else:
                  raise e
          os.symlink(target, pubname)

    remove_extraneous(self.pub_dir, expected_paths, self.ignored)
